
/* ckit.g */

/* --------------------------------------------------------------------- */

/*
literals // identifiers for C++ literals
{
   LPAREN     : '(' ;
   RPAREN     : ')' ;
   LBRACKET   : '[' ;
   RBRACKET   : ']' ;
   LBRACE     : '{' ;
   RBRACE     : '}' ;

   DOT        : '.'  ;
   ARROW      : "->" ;
   SCOPE      : "::" ;

   LOG_NOT    : '!' ;
   BIT_NOT    : '~' ;

   INC        : "++"  ;
   DEC        : "--"  ;

   DOT_STAR   : ".*"  ;
   ARROW_STAR : "->*" ;

   STAR       : '*' ;
   SLASH      : '/' ;
   MOD        : '%' ;

   PLUS       : '+' ;
   MINUS      : '-' ;

   SHL        : "<<" ;
   SHR        : ">>" ;

   LESS             : '<'  ;
   GREATER          : '>'  ;
   LESS_OR_EQUAL    : "<=" ;
   GREATER_OR_EQUAL : ">=" ;

   EQUAL            : "==" ;
   UNEQUAL          : "!=" ;

   BIT_AND    : '&' ;
   BIT_XOR    : '^' ;
   BIT_OR     : '|' ;

   LOG_AND    : "&&" ;
   LOG_OR     : "||" ;

   ASSIGN         : '='   ;
   MUL_ASSIGN     : "*="  ;
   DIV_ASSIGN     : "/="  ;
   MOD_ASSIGN     : "%="  ;
   PLUS_ASSIGN    : "+="  ;
   MINUS_ASSIGN   : "-="  ;
   SHL_ASSIGN     : "<<=" ;
   SHR_ASSIGN     : ">>=" ;
   BIT_AND_ASSIGN : "&="  ;
   BIT_XOR_ASSIGN : "^="  ;
   BIT_OR_ASSIGN  : "|="  ;

   QUESTION   : '?' ;
   COMMA      : ',' ;

   COLON      : ':' ;
   SEMICOLON  : ';' ;
   DOTS       : "..." ;
}
*/

/* --------------------------------------------------------------------- */

< struct CmmExpr { enum kind; } >
< struct CmmBinaryExpr : CmmExpr {  } >
< struct CmmStat { enum stat; } >

/*
identifier
number
real_number
char_literal
string_literal
*/

/* --------------------------------------------------------------------- */

/* compound name */

base_name <modify CmmName>:
  (
     spec_func: special_function
  |
     id: identifier
  );

base_args <modify CmmName>:
  (
    /* { is_template_arg }? */
      template_args:template_arg_list
  )? ;

cont_item <CmmContName>:
  <no_space> "::" <no_space>
  <modify> cont_name
  <execute add_name>
  <modify> cont_args ;

cont_name <modify CmmContName>:
  (
     '~' <no_space> <set a_destructor = true>
  |
     "template" <set a_template = true>
  )?
  (
     spec_func: special_function
  |
     id: identifier
  );

cont_args <modify CmmContName>:
  (
     /* { is_cont_template_arg }? */
        template_args:template_arg_list
  )? ;

compound_name <modify CmmName>:
  <execute open_name>
  <modify> base_name
  <execute start_name>
  <modify> base_args
  ( <add> cont_item )*
  <execute close_name>
  ;

simple_name <modify CmmName>:
  <execute open_name>
  <modify> base_name
  <execute start_name>
  <execute close_name>
  ;

global_mark <modify CmmName>:
  ( "::" <set a_global = true> )? ;

/* --------------------------------------------------------------------- */

/* qualified name */

qualified_name <CmmName>:
  // identifer used in simple expression
  <modify> compound_name ;

/* global name */

global_name <CmmName>:
  // namespaces, using declarations and class member initialization
  <modify> global_mark
  <modify> compound_name ;

/* declaring name */

global_declaring_name <CmmName>:
  // type used in declaration specifiers
  <modify> global_mark
  <modify> compound_name
  <execute verify_type>
  ;

/* type name */

global_type_name <CmmName>:
  // after typename keyword or in base class specification
  <execute open_type_name>
  <modify> global_mark
  <modify> compound_name
  <execute close_type_name>
  <execute verify_type>
  ;

/* new name */

simple_new_name <CmmName>:
  // namespace definition, enumeration item, template type parameter
  <execute open_new_name>
  <modify> simple_name
  <execute close_new_name>
  ;

global_new_name <CmmName>:
  // identifier in declarator, class or enumeration type
  <execute open_new_name>
  <modify> global_mark
  <modify> compound_name
  <execute close_new_name>
  ;

/* --------------------------------------------------------------------- */

/* identifier expr */

ident_expr <CmmIdentExpr:CmmExpr, kind=identExp>:
   name:qualified_name;

/* primary expression */

primary_expr <select CmmExpr>:
    numeric_literal_expr
  | real_literal_expr
  | char_literal_expr
  | string_literal_expr
  | this_expr
  | subexpr_expr
  | ident_expr // !? last - important
  ;

numeric_literal_expr <CmmNumValue:CmmExpr, kind=numValueExp>:
  value:number;

real_literal_expr <CmmRealValue:CmmExpr, kind=realValueExp>:
  value:real_number;

char_literal_expr <CmmCharValue:CmmExpr, kind=charValueExp>:
  value:character_literal;

string_literal_cont <CmmStringCont>:
  value:string_literal;

string_literal_expr <CmmStringValue:CmmExpr, kind=stringValueExp>:
  value:string_literal
  ( <add> string_literal_cont )* ; // !?

this_expr <CmmThisExpr:CmmExpr, kind=thisExp>:
  "this";

subexpr_expr <CmmSubexprExpr:CmmExpr, kind=subexprExp>:
  '(' param:expr ')';

/* postfix expression */

postfix_start <select CmmExpr>:
    modern_cast_expr
  | typeid_expr
  | typename_expr // !?
  | type_change_expr // !?
  | primary_expr ; // !? last - important

modern_cast_expr <CmmModernCastExpr:CmmExpr>:
  ( "dynamic_cast"       <set kind = DynamicCast>
  | "static_cast"        <set kind = StaticCast>
  | "const_cast"         <set kind = ConstCast>
  | "reinterpret_cast"   <set kind = ReinterpreterCast> )
  '<' type:type_id '>'
  '(' param:expr ')'
  ;

typeid_expr <CmmTypeidExpr:CmmExpr, kind=typeidExp>:
  "typeid"
  '('
  (
     { is_type () }?
        type:type_id
  |
     value:expr
  )
  ')'
  ;

typename_expr <CmmTypenameExpr:CmmExpr, kind=typenameExp>:
  "typename"
  name:global_type_name
  '(' list:expr_list ')' ;

type_change_expr <CmmTypechangeExpr:CmmExpr, kind=typechangeExp>: // !?
  (
    "signed"   <set a_signed = true>   |
    "unsigned" <set a_unsigned = true> |

    "short"    <set a_short=true>      |
    "long"     <set a_long=true>       ( "long" <set a_long_long = true> )? |

    "bool"     <set a_bool = true>     |
    "char"     <set a_char = true>     |
    "wchar_t"  <set a_wchar = true>    |
    "int"      <set a_int = true>      |
    "float"    <set a_float = true>    |
    "double"   <set a_double = true>   |
    "void"     <set a_void = true>
  )
  '(' list:expr_list ')' ;

postfix_expr <choose CmmExpr>:
  postfix_start
  (
    index_expr |
    call_expr |
    field_expr |
    ptr_field_expr |
    post_inc_expr |
    post_dec_expr
  )* ;

index_expr <CmmIndexExpr:CmmExpr, kind=indexExp>:
  <store left:CmmExpr>
  '[' right:expr ']' ;

call_expr <CmmCallExpr:CmmExpr, kind=callExp>:
  <store param:CmmExpr>
  '(' list:expr_list ')' ;

field_expr <CmmFieldExpr:CmmExpr, kind=fieldExp>:
  <store param:CmmExpr>
  <no_space> '.' <no_space>
  (
     '~' <no_space> <set destructor_name = true>
  |
     "template" <set template_name = true>
  )?
  <execute open_field_template>
  name:qualified_name
  <execute close_field_template> ;

ptr_field_expr <CmmPtrFieldExpr:CmmExpr, kind=ptrFieldExp>:
  <store param:CmmExpr>
  <no_space> "->" <no_space>
  (
     '~' <no_space> <set destructor_name = true>
  |
     "template" <set template_name = true>
  )?
  <execute open_ptr_field_template>
  name:qualified_name
  <execute close_ptr_field_template> ;

post_inc_expr <CmmPostIncExpr:CmmExpr, kind=postIncExp>:
  <store param:CmmExpr>
  "++" ;

post_dec_expr <CmmPostDecExpr:CmmExpr, kind=postDecExp>:
  <store param:CmmExpr>
  "--" ;

expr_list <CmmExprList>:
  (
     <add> assignment_expr
     ( ',' <add> assignment_expr )*
  )? ;

/* unary expression */

unary_expr <select CmmExpr>:
    inc_expr
  | dec_expr
  | deref_expr
  | addr_expr
  | plus_expr
  | minus_expr
  | bit_not_expr
  | log_not_expr
  | scope_expr
  | sizeof_expr
  | allocation_expr
  | deallocation_expr
  | throw_expr
  | postfix_expr ; // !? last - important

inc_expr <CmmUnaryExpr:CmmExpr, kind=incExp>:
  "++" cast_expr;

dec_expr <CmmUnaryExpr:CmmExpr, kind=decExp>:
  "--" param:cast_expr;

deref_expr <CmmUnaryExpr:CmmExpr, kind=derefExp>:
  '*' param:cast_expr;

addr_expr <CmmUnaryExpr:CmmExpr, kind=addrExp>:
  '&' param:cast_expr;

plus_expr <CmmUnaryExpr:CmmExpr, kind=plusExp>:
  '+' param:cast_expr;

minus_expr <CmmUnaryExpr:CmmExpr, kind=minusExp>:
  '-' param:cast_expr;

bit_not_expr <CmmUnaryExpr:CmmExpr, kind=bitNotExp>:
  '~' param:cast_expr;

log_not_expr <CmmUnaryExpr:CmmExpr, kind=logNotExp>:
  '!' param:cast_expr;

scope_expr <CmmUnaryExpr:CmmExpr, kind=scopeExp>: // !?
   "::" param:cast_expr ;

sizeof_expr <CmmSizeofExp:CmmExpr, kind=sizeofExp>:
  "sizeof"
  (
    { is_type_in_parenthesis () }?
       '(' type:type_id ')'
  |
     value:unary_expr
  );

/* new */

allocation_expr <CmmNewExpr:CmmExpr, kind=newExp>:
  // ( "::" <set a_global = true> )?
  "new"
  (
    { is_parenthesis_and_expression () }?
       placement:new_placement
  )?
  (
     type1:new_type_id
  |
    '(' type2:type_id ')'
  )
  (
    '(' init:expr_list ')'
  )? ;

new_placement <CmmNewPlacement>:
  '(' list:expr_list ')' ;

new_type_id <CmmNewTypeId>:
  type_spec:type_specifiers
  ptr:ptr_specifier_list
  ( <add> allocation_array_limit )* ;

allocation_array_limit <CmmNewArrayLimit>:
  '[' value:expr ']'  ;

/* delete */

deallocation_expr <CmmDeleteExpr:CmmExpr, kind=deleteExp>:
  // ( "::" <set a_global = true> )?
  "delete"
  ( '[' ']' <set a_array = true> )?
  param:cast_expr ;

/* cast expression */

cast_expr <select CmmExpr>:
    { is_type_in_parenthesis () }?
       cast_formula
  |
    unary_expr ;

cast_formula <CmmCastFormula: CmmExpr>:
  '(' type:type_id ')' param:cast_expr ;

/* pm expression */

pm_expr <choose CmmExpr>:
  cast_expr
  (
    <new CmmPtrMemExpr:CmmExpr>
    <store left:CmmExpr>
    ( ".*"  <set kind=dotMemberExp> |
      "->*" <set kind=arrowMemberExp> )
    right:cast_expr
  )* ;

/* arithmetic and logical expressions */

multiplicative_expr <choose CmmExpr>:
  pm_expr
  (
    <new CmmMulExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( '*' <set kind=mulExp> |
      '/' <set kind=divExp> |
      '%' <set kind=modExp> )
    right:pm_expr
  )* ;

additive_expr <choose CmmExpr>:
  multiplicative_expr
  (
    <new CmmAddExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( '+' <set kind=addExp> |
      '-' <set kind=subExp> )
    right:multiplicative_expr
  )* ;

shift_expr <choose CmmExpr>:
  additive_expr
  (
    <new CmmShiftExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( "<<" <set kind=shlExp> |
      ">>" <set kind=shrExp> )
    right:additive_expr
  )* ;

relational_expr <choose CmmExpr>:
  shift_expr
  (
    <new CmmRelExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( '<'  <set kind=ltExp> |
      '>'  <set kind=gtExp> |
      "<=" <set kind=leExp> |
      ">=" <set kind=geExp> )
    right:shift_expr
  )* ;

equality_expr <choose CmmExpr>:
  relational_expr
  (
    <new CmmEqExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( "==" <set kind=eqExp> |
      "!=" <set kind=neExp> )
    right:relational_expr
  )* ;

and_expr <choose CmmExpr>:
  equality_expr
  (
    <new CmmAndExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    '&' <set kind=bitAndExp>
    right:equality_expr
  )* ;

exclusive_or_expr <choose CmmExpr>:
  and_expr
  (
    <new CmmXorExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    '^' <set kind=bitXorExp>
    right:and_expr
  )* ;

inclusive_or_expr <choose CmmExpr>:
  exclusive_or_expr
  (
    <new CmmOrExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( '|' <set kind=bitOrExp> )
    right:exclusive_or_expr
  )* ;

logical_and_expr <choose CmmExpr>:
  inclusive_or_expr
  (
    <new CmmAndAndExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    "&&" <set kind=logAndExp>
    right:inclusive_or_expr
  )* ;

logical_or_expr <choose CmmExpr>:
  logical_and_expr
  (
    <new CmmOrOrExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    "||" <set kind=logOrExp>
    right:logical_and_expr
  )* ;

/* conditional and assignment expression */

assignment_expr <choose CmmExpr>:
  logical_or_expr // !?
  (
    <new CmmAssignExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( '='   <set kind=assignExp>    |
      "+="  <set kind=addAssignExp> |
      "-="  <set kind=subAssignExp> |
      "*="  <set kind=mulAssignExp> |
      "/="  <set kind=divAssignExp> |
      "%="  <set kind=modAssignExp> |
      "<<=" <set kind=shlAssignExp> |
      ">>=" <set kind=shrAssignExp> |
      "&="  <set kind=andAssignExp> |
      "^="  <set kind=xorAssignExp> |
      "|="  <set kind=orAssignExp>  |
      '?'   <set kind=condExp> middle:expr ':' )
    right:logical_or_expr // !?
  )? ;

conditional_expr <return CmmExpr>:
  assignment_expr ;

/* expression */

expr <choose CmmExpr>:
  assignment_expr
  (
    <new CmmCommaExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( ',' <set kind=commaExp> )
    right:assignment_expr
  )* ;

/* constant expression */

const_expr <return CmmExpr>:
  conditional_expr ;

/* --------------------------------------------------------------------- */

/* statement */

stat <select CmmStat>:

   ( identifier ':' ) => labeled_stat |
   { is_declaration () }? declaration_stat |
   expression_stat |
   empty_stat |

   compound_stat |
   case_stat |
   default_stat |
   if_stat |
   switch_stat |
   while_stat |
   do_stat |
   for_stat |
   break_stat |
   continue_stat |
   return_stat |
   goto_stat |
   try_stat;

/* inner_stat - statements with different output indentation */

inner_stat <return CmmStat>:
   <new_line>
   <indent>
   stat
   <unindent>;

/* list of statements */

stat_list <CmmStatSect>:
   ( <new_line> <add> stat )* ;

/* conditions */

condition <CmmCondition>:
     { is_declaration () }?
        cond_decl:condition_declaration
  |
     cond_expr:expr
  ;

/* concrete statements */

declaration_stat <CmmDeclStat:CmmStat, stat=declarationStat>:
  inner_decl:block_declaration ;

labeled_stat <CmmLabeledStat:CmmStat, stat=labeledStat>:
  lab:identifier ':'
  <new_line>
  body:stat ;

case_stat <CmmCaseStat:CmmStat, stat=caseStat>:
  "case" case_expr:expr ':'
  <new_line>
  body:stat;

default_stat <CmmDefaultStat:CmmStat, stat=defaultStat>:
  "default" ':'
  <new_line>
  body:stat;

expression_stat <CmmExprStat:CmmStat, stat=expressionStat>:
  inner_expr:expr ';';

empty_stat <CmmEmptyStat:CmmStat, stat=emptyStat>:
  ';' ;

compound_stat <CmmCompoundStat:CmmStat, stat=compoundStat>:
  <execute open_block >
  '{'
  <indent>
  body:stat_list
  <unindent>
  '}'
  <execute close_block > ;

if_stat <CmmIfStat:CmmStat, stat=ifStat>:
  <execute open_block >
  "if" '(' cond:condition ')' then_stat:inner_stat
  (
    <silent>
    <new_line>
    "else" else_stat:inner_stat
  )?
  <execute close_block > ;

switch_stat <CmmSwitchStat:CmmStat, stat=switchStat>:
  <execute open_block >
  "switch" '(' cond:condition ')' body:inner_stat
  <execute close_block > ;

while_stat <CmmWhileStat:CmmStat, stat=whileStat>:
  <execute open_block >
  "while" '(' cond:condition ')' body:inner_stat
  <execute close_block > ;

do_stat <CmmDoStat:CmmStat, stat=doStat>:
  <execute open_block >
  "do" body:inner_stat "while" '(' cond_expr:expr ')' ';'
  <execute close_block > ;

for_stat <CmmForStat:CmmStat, stat=forStat>:
  <execute open_block >
  "for"
  '('
  ( from_expr:condition )?
  ';'
  (to_expr:expr)?
  ';'
  (cond_expr:expr)?
  ')'
  body:inner_stat
  <execute close_block > ;

break_stat <CmmBreakStat:CmmStat, stat=breakStat>:
  "break" ';' ;

continue_stat <CmmContinueStat:CmmStat, stat=continueStat>:
  "continue" ';' ;

return_stat <CmmReturnStat:CmmStat, stat=returnStat>:
  "return" (return_expr:expr)? ';' ;

goto_stat <CmmGotoStat:CmmStat, stat=gotoStat>:
  "goto" goto_lab:identifier ';' ;

/* ---------------------------------------------------------------------- */

/* declaration */

declaration_list <CmmDeclSect>:
  ( <add> declaration <new_line> )* ;

declaration <select CmmDecl>:

    ( "extern" string_literal ) => linkage_declaration
  | ( "extern" "template" ) => extern_template_declaration // !? GNU extension
  | /* { is_constructor () }? */ (constructor_declaration) => constructor_declaration
  | simple_declaration
  | empty_declaration

  | namespace_declaration
  | using_declaration
  | asm_declaration

  | template_declaration ;

block_declaration <select CmmDecl>: // used only in declaration_stat

    local_declaration

  | namespace_declaration
  | using_declaration
  | asm_declaration ;

/* ---------------------------------------------------------------------- */

/* namespace declaration */

namespace_declaration <CmmNamespaceDecl:CmmDecl> :
  "namespace"
  (
     name:simple_new_name
     (
        '=' value:global_name ';'
        <execute add_namespace_alias>
     |
        <modify> namespace_body
     )
  |
     <modify> namespace_body
  );

namespace_body <modify CmmNamespaceDecl> :
   <new_line>
   <execute open_namespace> // before brace, important for include(s) on lexical level
   '{'
      <indent>
      body:declaration_list
      <unindent>
   <execute close_namespace> // before brace, important for include(s)
   '}'
   ;

/* using declaration */

using_declaration <CmmUsingDecl:CmmDecl>:
  "using"
  (
     "namespace" <set a_namespace = true>
  |
     "typename"  <set a_typename  = true>
  )?
  <execute open_using>
  name: global_name
  <execute close_using>
  ';' ;

/* linkage declaration */

linkage_declaration <CmmExternDecl:CmmDecl>:
  "extern"
  language:string_literal
  (
     '{' decl_list:declaration_list '}'
  |
     inner_decl:declaration
  ) ;

/* asm declaration */

asm_declaration <CmmAsmDecl:CmmDecl>:
  "asm" '(' instruction:string_literal ')' ';' ;

/* empty declaration */

empty_declaration <CmmEmptyDecl:CmmDecl>:
  ';' ;

/* ---------------------------------------------------------------------- */

/* declaration specifiers */

decl_specifiers <CmmDeclSpec>:
  <execute open_decl_specifiers >
  <modify> simple_decl_specifiers
  (
     basic_name:global_declaring_name
  |
     class_spec:class_specifier
  |
     enum_spec:enum_specifier
  |
     typename_spec:typename_specifier
  |
     <modify> simple_type_specifiers
  )
  <modify> simple_decl_specifiers // !?
  <execute close_decl_specifiers>
  ;

type_specifiers <CmmDeclSpec>:
  <modify> simple_cv_specifiers
  (
    "register" <set a_register = true> // !? root/cint/inc/C__ci.h
  )?
  (
     basic_name:global_declaring_name // !?
  |
     class_spec:elaborated_class_specifier
  |
     enum_spec:elaborated_enum_specifier
  |
     typename_spec:typename_specifier
  |
     <modify> simple_type_specifiers
  )
  <modify> simple_decl_specifiers // !?
  ;

typename_specifier <CmmTypenameSpec>:
  "typename"
  name:global_type_name ;

simple_decl_specifiers <modify CmmDeclSpec>:
  (
    "typedef"  <set a_typedef = true>  |
    "friend"   <set a_friend = true>   |

    "inline"   <set a_inline = true>   |
    "virtual"  <set a_virtual = true>  |
    "explicit" <set a_explicit = true> |

    "mutable"  <set a_mutable = true>  |
    "extern"   <set a_extern = true>   |
    "static"   <set a_static = true>   |
    "auto"     <set a_auto = true>     |
    "register" <set a_register = true> |

    "const"    <set a_const = true>    |
    "volatile" <set a_volatile = true>
  )* ;

constructor_decl_specifiers <CmmDeclSpec>:
  ( "inline"   <set a_inline = true>   |
    "virtual"  <set a_virtual = true>  |
    "explicit" <set a_explicit = true>
  )* ;

simple_cv_specifiers <modify CmmDeclSpec>:
  (
    "const" <set a_const = true> |
    "volatile" <set a_volatile = true>
  )* ;

simple_type_specifiers <modify CmmDeclSpec>:
  ( "signed"   <set a_signed = true>   |
    "unsigned" <set a_unsigned = true> |

    "short"    <set a_short = true>    |
    "long"     <set a_long = true>     ( "long" <set a_long_long = true> )? |

    "bool"     <set a_bool = true>     |
    "char"     <set a_char = true>     |
    "wchar_t"  <set a_wchar = true>    |
    "int"      <set a_int = true>      |
    "float"    <set a_float = true>    |
    "double"   <set a_double = true>   |
    "void"     <set a_void = true>
  )@ ;

/* ---------------------------------------------------------------------- */

/* pointer specifiers */

ptr_specifier_list <CmmPtrSpecifierSect>:
  ( <add> ptr_specifier )*
  <execute member_pointer_list> // !?
  ;

ptr_specifier <CmmPtrSpecifier>:
    '*' <set pointer=true>
    <modify> ptr_cv_specifier_list
  |
    '&' <set reference=true>
    <modify> ptr_cv_specifier_list
  ;

member_specifier <CmmPtrSpecifier>: // !? only for output
  area:global_name
  <modify> member_specifier_tail ;

member_specifier_tail <modify CmmPtrSpecifier>:
  "::" '*'
  <set member_pointer=true>
  <modify> ptr_cv_specifier_list ;

ptr_cv_specifier_list <modify CmmPtrSpecifier>:
  (
    "const" <set cv_const = true>
  |
    "volatile" <set cv_volatile = true>
  )* ;

/* function and array specifiers */

cont_specifier_list <CmmContSpecifierSect>:
  ( <add> cont_specifier )* ;

cont_specifier <select CmmContSpecifier>:
    function_specifier
  |
    array_specifier ;

function_specifier <CmmFunctionSpecifier:CmmContSpecifier>:
  <execute open_parameters>
  '('
  parameters:parameter_declaration_list
  ')'
  (
    "const" <set cv_const = true>
  |
    "volatile" <set cv_volatile = true>
  )*
  ( exception_spec:exception_specification )?
  <execute close_parameters> ;

array_specifier <CmmArraySpecifier:CmmContSpecifier>:
  '['
  (lim:expr)?
  ']' ;

/* parameter declarations */

parameter_declaration_list <CmmParamSect>:
  (
     <add> parameter_declaration
     ( ',' <add> parameter_declaration )*
  )? ;

parameter_declaration <CmmParamItem> :
  (
     { is_parameter () }?
       type:common_type_id
       <modify> parameter_initialization
       ( "..." <set following_dots=true> )? // !?
  |
     { is_initializer_acceptable () }?
     value:expr
  |
     "..." <set dots=true> // !?
  )
  <execute add_parameter>
  ;

parameter_initialization <modify CmmParamItem>:
  ( '=' init:assignment_expr )? ;

/* ---------------------------------------------------------------------- */

// simple_declaration      ... used in declaration (global) and member_item (class)
// constructor_declaration ... used in declaration (global) and member_item (class)
// local_declaration       ... used in declaration_statement
// condition_declaration   ... used in "if", "switch", "while", "for" statements

// simple_declaration:
//    with decl_specifiers
//    possible without declarator
//       necessary for class declarations without init declarator
//       i.e: class C { ... };
//    possible with function body

// local_declaration:
//    with decl_specifiers
//    with at least one declarator

// constructor_declaration:
//    without decl_specifiers ( inline, virtual, explicit is acceptable )
//    with at least one declarator

/* ---------------------------------------------------------------------- */

/* simple declaration */

simple_declaration <CmmSimpleDecl:CmmDecl>:
  spec:decl_specifiers
  <execute open_simple_declaration>
  (
      // ':' empty_width:conditional_expr ';' // !?
  // |
     ';' <set empty=true> // !?
  |
  <add> init_declarator
  (
     func_body:function_body
  |
        ( ',' <add> init_declarator )*
     ';'
  )
  )
  <execute close_simple_declaration>
  ;

/* local declaration */

local_declaration <CmmSimpleDecl>:
  spec:decl_specifiers
  <execute open_simple_declaration>
  <add> init_declarator
  ( ',' <add> init_declarator )*
  ';'
  <execute close_simple_declaration>
  ;

/* condition declaration */

condition_declaration <CmmSimpleDecl>:
  spec:decl_specifiers
  <execute open_simple_declaration>
  <add> init_declarator
  ( ',' <add> init_declarator )*
  // no semicolon
  <execute close_simple_declaration>
  ;

/* constructor declaration */

constructor_declaration <CmmCtorDecl:CmmSimpleDecl>:
  <execute open_constructor_declaration>
  spec:constructor_decl_specifiers
  ( '~' <set destructor=true> )?
  /* <add> init_declarator */
  <add> declarator
  (
     func_body:function_body
  |
     ';' // no empty = true
  )
  <execute close_constructor_declaration>
  ;

/* constructor declaration head - only for predicate */

constructor_head <CmmCtorHead:CmmCtorDecl>:
  spec:constructor_decl_specifiers
  ( '~' <set destructor=true> )?
  <add> constructor_head_declarator
  (
     ':' <set init_values = true>
  |
     '{' <set body=true>
  |
     "try" <set try_block = true>
  |
     "=" <set equal = true>
  |
     ';' // no empty = true
  ) ;

constructor_head_declarator <CmmDeclarator>:
  <execute open_declarator>
  name:global_new_name
  cont:constructor_head_specifier_list
  <execute close_declarator>
  ;

constructor_head_specifier_list <CmmContSpecifierSect>:
  <add> constructor_head_specifier;

constructor_head_specifier <return CmmContSpecifier>:
  function_specifier ;

/* ---------------------------------------------------------------------- */

/* function body */

function_body <CmmFunctionBody>:
  <execute open_function>
  (
     ( ':' <new_line> ctor_init:ctor_initializer )?
     <new_line>
     statements:compound_stat
  |
     <new_line>
     "try"
     ( ':' ctor_init:ctor_initializer )?
     statements:compound_stat
     handlers:handler_list
  )
  <execute close_function>
  ;

/* ---------------------------------------------------------------------- */

// declarator          ... used in init_declarator
// abstract_declarator ... used in type_id
// common_declarator   ... used in common_type_id

// declarator          ... name or inner declarator
// abstract_declarator ... empty or inner declarator
// common_declarator   ... empty, name or inner declarator

// declarator:
//   parenthesis allowed only around real nested declarator
//   otherwise function parameters or variable initializers are miss understood

/* ---------------------------------------------------------------------- */

/* declarator */

declarator <CmmDeclarator>:
  <execute open_declarator>
  ptr:ptr_specifier_list
  (
     name:global_new_name
  |
     '(' inner:declarator ')'
  )
  cont:cont_specifier_list
  <execute close_declarator>
  ;

abstract_declarator <CmmDeclarator>:
  <execute open_declarator>
  ptr:ptr_specifier_list
  (
     { is_inner_abstract_declarator () }?
        '(' inner:abstract_declarator ')'
  )?
  cont:cont_specifier_list
  <execute close_declarator>
  ;

common_declarator_name <modify CmmDeclarator>:
  name:global_new_name
  ;

common_declarator <CmmDeclarator>:
  <execute open_declarator>
  ptr:ptr_specifier_list
  (
     { is_inner_common_declarator () }?
        '(' inner:common_declarator ')'
  |
     <modify> common_declarator_name
     <set empty_branch = true>
  )?
  cont:cont_specifier_list
  <execute close_declarator>
  ;

/* ---------------------------------------------------------------------- */

// type_id ... used in expressions, initialization of template_type_param, template_type_arg, exception_specification
// common_type_id ... used in parameter_declaration, template_value_param, try handler

// init_declarator ... used in simple_declaration, constructor_declaration, local_declaration

/* ---------------------------------------------------------------------- */

/* type id */

type_id <CmmTypeId>: /* abstract_type_id */
  spec:type_specifiers
  decl:abstract_declarator ;

common_type_id <CmmTypeId>:
  spec:type_specifiers
  decl:common_declarator ;

/* init declarator */

init_declarator <return CmmDeclarator>:
  declarator
  <execute add_init_declarator>
  ( /* { is_width_acceptable }? */  ( ':' conditional_expr (',' | ';') ) =>  ':' width:conditional_expr )? // !?
  ( init:initializer  )? ;

/* initializer */

initializer <CmmInitializer>:
   '=' value:initializer_item
   // | "(" value2:expr ")" // !?
   ;

initializer_item <select CmmInitItem>:
     simple_initializer
   | initializer_list ;

empty_initializer <CmmInitEmpty:CmmInitItem>:
   ;

simple_initializer <CmmInitSimple:CmmInitItem>:
     ('}') => <set empty = true>
  |
     expr:assignment_expr ;

initializer_list <CmmInitList:CmmInitItem>:
  '{'
  <indent>
  (
    <add> initializer_item
    (
      ','
      <new_line>
      <add> initializer_item
    )*
  )?
  <unindent>
  '}';

/* ---------------------------------------------------------------------- */

/* elaborated type specifiers */

elaborated_enum_specifier <CmmEnumDecl>:
  "enum"
  name:global_new_name ; // !?

elaborated_class_specifier <CmmClassDecl>:
  (
    "class"  <set style = ClassStyle> |
    "struct" <set style = StructStyle> |
    "union"  <set style = UnionStyle>
  )
  name:global_new_name; // !?

/* enum and class declarations */

class_declaration <return CmmClassDecl> :
  class_specifier ';' ;

enum_declaration <return CmmEnumDecl>:
  enum_specifier  ';' ;

/* ---------------------------------------------------------------------- */

/* enum */

enum_specifier <CmmEnumDecl:CmmDecl>:
  "enum"
  (
     name:global_new_name
     (
        <modify> enum_body
        <set with_body = true>
     |
        <execute add_enum_forw>
        // NO <set forward = true>
     )
  |
     <modify> enum_body
     // NO <set anonymous = true>
  );

enum_body <modify CmmEnumDecl>:
  <execute open_enum>
  <new_line>
  '{'
     <indent>
     enum_items:enum_list
     <unindent>
  '}'
  <execute close_enum>
  ;

enum_list <CmmEnumSect>:
  (
    <add> enumerator
    ( ',' <add> enumerator )*
  )? ;

enumerator <CmmEnumItem>:
     ( '}' ) => <set empty=true>
  |
     <new_line>
     name:simple_new_name
     ( '=' init:const_expr )?
     <execute add_enum_item> ;

/* ---------------------------------------------------------------------- */

/* class */

class_specifier <CmmClassDecl:CmmDecl>:
  (
    "class"  <set style = ClassStyle> |
    "struct" <set style = StructStyle> |
    "union"  <set style = UnionStyle>
  )
  (
    name:global_new_name
    (
      <silent>
      <modify> class_body
      <set with_body = true>
    |
      // NO <set forward = true>
      <execute add_class_forw>
    )
  |
    <modify> class_body
    // NO <set anonymous = true>
  );

class_body <modify CmmClassDecl>:
  ( ':' base_list:base_specifier_list )?
  <execute open_class>
  <new_line>
  '{'
     <indent>
     members:member_list
     <unindent>
  '}'
  <execute close_class>
  ;

/* members */

member_visibility <CmmMemberVisibility:CmmDecl>:
  ( "private"   <set access = PrivateAccess>
  | "protected" <set access = ProtectedAccess>
  | "public"    <set access = PublicAccess> )
  <no_space> ':'
  <execute add_visibility> ;

member_item <select CmmDecl>:

    member_visibility

  | ("explicit") => constructor_declaration
  | /* { is_constructor () }? */ (constructor_declaration) => constructor_declaration
  | simple_declaration // !?
  | empty_declaration

  | using_declaration
  | template_declaration
  ;

member_list <CmmDeclSect>:
  ( <add> member_item <new_line> )*
  ;

/* base specification */

base_specifier_list <CmmBaseSect>:
  <add> base_specifier
  ( ',' <add> base_specifier )* ;

base_specifier <CmmBaseItem>:
  <modify> virtual_base_specifier
  ( "private"   <set access = PrivateAccess>
  | "protected" <set access = ProtectedAccess>
  | "public"    <set access = PublicAccess>
  |             <set access = NoAccess> )
  <modify> virtual_base_specifier
  from_cls:global_type_name
  <execute add_base> ;

virtual_base_specifier <modify CmmBaseItem>:
  ( "virtual" <set a_virtual = true> )? ;

/* constructor initializer */

ctor_initializer <CmmCtorInitializer>:
  <indent>
  <add> member_initializer
  (
     ','
     <new_line>
     <add> member_initializer
  )*
  <unindent> ;

member_initializer <CmmMemberInitializer>:
   name:global_name // !? global_name
   '('
   params:expr_list
   ')';

/* ---------------------------------------------------------------------- */

/* special member functions */

conversion_specifiers <return CmmDeclSpec>: // !?
  /* <execute open_conversion > */
  type_specifiers
  ptr:ptr_specifier_list
  /* <execute close_conversion > */ ;

special_function <CmmSpecialFuction>:
   "operator"
   (
     conv:conversion_specifiers // !?

   | "new"    <set spec_new=true>    ( ('[' ']') => '[' ']' <set spec_new_array=true> )? // !?
   | "delete" <set spec_delete=true> ( ('[' ']') => '[' ']' <set spec_delete_array=true> )?

   | "+" <set spec_add=true>
   | "-" <set spec_sub=true>
   | "*" <set spec_mul=true>
   | "/" <set spec_div=true>
   | "%" <set spec_mod=true>
   | "^" <set spec_xor=true>
   | "&" <set spec_and=true>
   | "|" <set spec_or=true>
   | "~" <set spec_not=true>

   | "!"  <set spec_log_not=true>
   | "="  <set spec_assign=true>
   | "<"  <set spec_lt=true>
   | ">"  <set spec_gt=true>
   | "+=" <set spec_add_assign=true>
   | "-=" <set spec_sub_assign=true>
   | "*=" <set spec_mul_assign=true>
   | "/=" <set spec_div_assign=true>
   | "%=" <set spec_mod_assign=true>

   | "^="  <set spec_xor_assign=true>
   | "&="  <set spec_and_assign=true>
   | "|="  <set spec_or_assign=true>
   | "<<"  <set spec_shl=true>
   | ">>"  <set spec_shr=true>
   | ">>=" <set spec_shl_assign=true>
   | "<<=" <set spec_shr_assign=true>
   | "=="  <set spec_eq=true>
   | "!="  <set spec_ne=true>

   | "<="  <set spec_le=true>
   | ">="  <set spec_ge=true>
   | "&&"  <set spec_log_and=true>
   | "||"  <set spec_log_or=true>
   | "++"  <set spec_inc=true>
   | "--"  <set spec_dec=true>
   | ","   <set spec_comma=true>
   | "->*" <set spec_member_deref=true>
   | "->"  <set spec_deref=true>

   | "(" ")" <set spec_call=true>
   | "[" "]" <set spec_index=true>
   ) ;

/* ---------------------------------------------------------------------- */

/* template declaration */

template_declaration <CmmTemplateDecl:CmmDecl>:
  <modify> common_template_declaration ;

extern_template_declaration <CmmExxternTemplateDecl:CmmTemplateDecl>:
  // GNU extension
  "extern"
  <modify> common_template_declaration ;

common_template_declaration <modify CmmTemplateDecl:CmmDecl>:
  <execute open_template>
  <modify> template_specification
  <execute add_template>
  inner_decl:declaration
  <execute close_template>
  ;

template_specification <modify CmmTemplateDecl>:
  (
    "export"
    <set export_template = true>
  )?
  "template"
  (
    '<'
    (
    params:template_param_list
    |
      <set explicit_specialization = true>
    )
    '>'
    <set parameter_block = true> // !?
  |
    <set explicit_instatiation = true>
  );

template_param_list <CmmTemplateParamSect>:
    <add> template_param
  ( ',' <add> template_param )* ;

template_param <select CmmTemplateParam>:
    ( "template" | "class" | "typename" ) => template_type_param
  |
    template_value_param
  ;

template_type_param <CmmTemplateTypeParam:CmmTemplateParam>:
  (
    "template" '<' params:template_param_list '>'
  )?
  (
    "class" <set a_class = true>
  |
    "typename" <set a_typename = true>
  )
  ( name:simple_new_name )?
  ( '=' init_type:type_id )?
  <execute add_template_type_param>
  ;

template_value_param <CmmTemplateValueParam:CmmTemplateParam>:
  type:common_type_id
  <execute add_template_value_param>
  ( '=' init_value:shift_expr )? ; // !?

/* template arguments */

template_arg_list <CmmTemplateArgSect>:
  <execute open_template_arguments>
  '<'
  (
    <add> template_arg
    ( ',' <add> template_arg )*
  )?
  '>'
  <execute close_template_arguments> ;

template_arg <select CmmTemplateArg>:
    /* { is_type () }? */
      template_type_arg
  |
    template_value_arg;

template_type_arg <CmmTemplateTypeArg:CmmTemplateArg>:
  type:type_id; // !?

template_value_arg <CmmTemplateValueArg:CmmTemplateArg>:
   value:shift_expr; // !?

/* ---------------------------------------------------------------------- */

/* try statement */

try_stat <CmmTryStat:CmmStat, stat=tryStat>:
  "try"
  <new_line>
  body:compound_stat
  handlers:handler_list ;

handler_list <CmmHandlerSect>:
  <add> handler
  ( <add> handler )* ;

handler <CmmHandlerItem>:
  <new_line>
  "catch"
  '('
  (
    type:common_type_id
  |
    "..." <set dots=true>
  )
  ')'
  <new_line>
  body:compound_stat ;

/* throw expression */

throw_expr <CmmThrowExpr:CmmExpr, kind=throwExp >:
  "throw" (expr:assignment_expr)? ;

/* exception specification */

exception_specification <CmmExceptionSect>:
  "throw"
  '('
  (
    <add> exception_specification_item
    (',' <add> exception_specification_item)*
  )?
  ')' ;

exception_specification_item <CmmExceptionItem: CmmTypeDecl>:
  type:type_id ;

/* ---------------------------------------------------------------------- */
